#include <linux/kernel.h>
#include <linux/init.h>

#include <asm/pgtable.h>

#include <asm/xen/hypercall.h>
#include <xen/page.h>
#include <xen/hvc-console.h>

#include "xen-ops.h"
#include "init.h"
#include "mmu.h"

void __meminit early_make_page_readonly(void *va, unsigned int feature)
{
	unsigned long addr, _va = (unsigned long)va;
	int r;
	unsigned long idx;
	pte_t pte, *ptep;
	unsigned long *page = (unsigned long *) init_level4_pgt;

	if (xen_feature(feature))
		return;

	idx = pgd_index(_va);

	addr = (unsigned long) page[idx];
	addr_to_page(addr, page);

	idx = pud_index(_va);
	addr = page[idx];

	addr_to_page(addr, page);

	idx = pmd_index(_va);
	addr = page[idx];
	addr_to_page(addr, page);

	idx =pte_index(_va);
	ptep = (pte_t *) &page[idx];

	pte.pte = ptep->pte & ~_PAGE_RW;

	if ( (r = HYPERVISOR_update_va_mapping(_va, pte, 0)) ) {
		xprintk("%s: failure: %d :(\n", __func__, r);
		BUG();
	}
}



void __init xen_init_pt(void)
{
	unsigned long addr, *page;

	xprintk("%s starting\n", __func__);

	xprintk("xen_cr3: %lx. xen_cur_cr3: %lx\n", x86_read_percpu(xen_cr3), x86_read_percpu(xen_current_cr3));

	/* Find the initial pte page that was built for us. */
	page = (unsigned long *)xen_start_info->pt_base;
	xprintk("page1: %p\n", page);
	addr = page[pgd_index(__START_KERNEL_map)];
	xprintk("addr1: %lx\n", addr);
	addr_to_page(addr, page);
	xprintk("addr1, page2: %lx, %p\n", addr, page);

	addr = page[pud_index(__START_KERNEL_map)];
	xprintk("addr2: %lx\n", addr);

	addr_to_page(addr, page);
	xprintk("addr2,page2: %lx, %p\n", addr, page);


#if 0  /* CONFIG_XEN_COMPAT <= 0x030002 */
	/* On Xen 3.0.2 and older we may need to explicitly specify _PAGE_USER
	   in kernel PTEs. We check that here. */
	if (HYPERVISOR_xen_version(XENVER_version, NULL) <= 0x30000) {
		unsigned long *pg;
		pte_t pte;

		/* Mess with the initial mapping of page 0. It's not needed. */
		BUILD_BUG_ON(__START_KERNEL <= __START_KERNEL_map);
		addr = page[pmd_index(__START_KERNEL_map)];
		addr_to_page(addr, pg);
		pte.pte = pg[pte_index(__START_KERNEL_map)];
		BUG_ON(!(pte.pte & _PAGE_PRESENT));

		/* If _PAGE_USER isn't set, we obviously do not need it. */
		if (pte.pte & _PAGE_USER) {
			/* _PAGE_USER is needed, but is it set implicitly? */
			pte.pte &= ~_PAGE_USER;
			if ((HYPERVISOR_update_va_mapping(__START_KERNEL_map,
							  pte, 0) != 0) ||
			    !(pg[pte_index(__START_KERNEL_map)] & _PAGE_USER))
				/* We need to explicitly specify _PAGE_USER. */
				__kernel_page_user = _PAGE_USER;
		}
	}
#endif

	xprintk("l4pgt: %p\n", init_level4_pgt);

	/* The page tables are pre-initialized with values that are not
	 * usable by Xen. Zero them out.
	 */
	memset(init_level4_pgt, 0, PAGE_SIZE);
	memset(level3_kernel_pgt, 0, PAGE_SIZE);
	memset(level2_kernel_pgt, 0, PAGE_SIZE);

	xprintk("%s going change init_level4_pgt\n", __func__);

	/* Construct mapping of initial pte page in our own directories. */
	init_level4_pgt[pgd_index(__START_KERNEL_map)] = 
		mk_kernel_pgd(__pa_symbol(level3_kernel_pgt));

	xprintk("pointed l4 to l3kpgt\n");

	xprintk("l3pgt: %p\n", level3_kernel_pgt);


	{
	unsigned long pi = pud_index(__START_KERNEL_map);
	pud_t newp;
	/*int i;
	for (i = 0; i < PTRS_PER_PUD; i++)
		xprintk("l3[%d] (%p)= %lx\n", i, &level3_kernel_pgt[i], (unsigned long)level3_kernel_pgt[i].pud);*/

	xprintk("pud index: %lx\n", pi);
	xprintk("writing to %p\n", &level3_kernel_pgt[pi]);
	newp = __pud(__pa_symbol(level2_kernel_pgt) |
		      _KERNPG_TABLE);
	xprintk("new pud: %lx\n", (unsigned long)newp.pud);
	level3_kernel_pgt[pi] = newp;

	}
	xprintk("pointed l3 to l2kpgt\n");

	xprintk("l2pgt: %p\n", level2_kernel_pgt);
	memcpy((void *)level2_kernel_pgt, page, PAGE_SIZE);
	xprintk("copied l2kpgt from page\n");


	xprintk("%s: going to pin\n", __func__);

	early_make_page_readonly(init_level4_user_pgt,
				 XENFEAT_writable_page_tables);
	early_make_page_readonly(level3_user_pgt,
				 XENFEAT_writable_page_tables);

	xprintk("%s going to make pages readonly\n", __func__);
	early_make_page_readonly(init_level4_pgt,
				 XENFEAT_writable_page_tables);
	early_make_page_readonly(level3_kernel_pgt,
				 XENFEAT_writable_page_tables);
	early_make_page_readonly(level2_kernel_pgt,
				 XENFEAT_writable_page_tables);

	xprintk("%s going to pin pgds\n", __func__);
	if (!xen_feature(XENFEAT_writable_page_tables)) {
		xprintk("kernel:\n");
		xen_do_pin(MMUEXT_PIN_L4_TABLE, PFN_DOWN(__pa_symbol(init_level4_pgt)));
		xprintk("user:\n");
		xen_do_pin(MMUEXT_PIN_L4_TABLE, PFN_DOWN(__pa_symbol(init_level4_user_pgt)));
	}


	/*FIXME: check why this isn't set before xen_pgd_pin() on
	 * x86_64 XenSource code
	 */
	xprintk("%s: set_pgd:\n", __func__);
	set_pgd((pgd_t *)(init_level4_user_pgt + 511), 
		mk_kernel_pgd(__pa_symbol(level3_user_pgt)));
	xprintk("%s: returning.\n", __func__);

	xprintk("going to load new pagetable:\n");
	write_cr3(__pa(init_level4_pgt));
	xprintk("loaded new pagetable\n");
}


